from math import *
import weakref

from engine.Hud import Hud
from engine import Randomness

from Robot import Robot
from BadRobot import BadRobot

from Ground import Ground
from Tree import Tree, treeAt
from Rock import Rock
from Stone import Stone

from Pod import Pod
from Food import Food
from Fish import Fish
from Disk import Disk

from Scenery import Scenery
from Item import Item
from Character import Character

class World:
	"""
	This entity is the game world, which holds all of the other entities (characters, etc.)
	It also manages the drawing of the background.
	"""
	def __init__(self):
		"""
		Sets up the world, creates a default character and the first screen, etc.
		"""
		# (how far away the horizon is in screen distances if we were looking top down)
		self.horizon = 2.0
		self.entities = []
		self.position = [0, 0]
		
		self.player = Robot(self)
		self.ground = Ground(self)
		self.hud = Hud(self, self.player)
		
		# loosely tracks where existant objects are in our world
		self.where = weakref.WeakValueDictionary()
		# remembers objects have been picked up so they don't get deleted from the where weakref
		self.pickedup = []
		# dead guys we have defeated - keep them around
		self.dead = []
		
		# where we last created a set of items
		self.lastcreated = [0, 0]
		
		# this is a list of Item entities that are near the player.
		
		self.entities.append(self.player)
		self.entities.append(self.ground)
		
		for x in range(-4, 4):
			for y in range(-8, 4):
				self.CreateStuff(x/4.0, y/4.0)
		
		# self.AddEntity(BadRobot(self, 1.5, 0.5), (1.5, 0.5))

	def CreateStuff(self, x, y):
		v = int(Randomness.Noise2D(x, y, "level") * 100)
		if v in [11, 61] and not self.TestPos(x,y):
			self.AddEntity(Tree(self, x, y), (x,y))
		if v in [2, 3, 4, 5, 6, 7, 8, 52, 53, 54, 55, 56, 57, 58] and not self.TestPos(x,y):
			self.AddEntity(Stone(self, x, y), (x,y))
		if v in [9, 59] and not self.TestPos(x,y):
			self.AddEntity(Pod(self, x, y), (x,y))
		if v in [10, 60] and not self.TestPos(x,y):
			self.AddEntity(Food(self, x, y), (x,y))
		if v in [15, 65] and not self.TestPos(x,y):
			self.AddEntity(Fish(self, x, y), (x,y))
		if v in [16, 66] and not self.TestPos(x,y):
			self.AddEntity(Disk(self, x, y), (x,y))
		if v == 50 and not self.TestPos(x, y):
			self.AddEntity(BadRobot(self, x, y), (x, y))
	
	def SetDead(self, who):
		self.dead.append(who)
	
	def TestPos(self, x, y):
		pos = (int(x * 4), int(y * 4))
		return self.where.has_key(pos)

	def AddEntity(self, e, where):
		"""
		Add an entity and update the weakref location for this entity so we can tell what we're tracking where.
		if the entity is already in our weakref tracker then delete it don't add it.
		"""
		pos = (int(where[0] * 4), int(where[1] * 4))
		if not self.where.has_key(pos):
			#print "created", e, pos
			self.where[pos] = e
			self.entities.append(e)
		else:
			#print "not creating", e, pos
			del e
		#print "entities: ", len(self.entities)
	
	def RemoveEntity(self, e):
		"""
		Remove an entity from our list and delete it.
		"""
		# don't get rid of dead guys
		if not e in self.dead:
			self.entities.remove(e)
			del e
	
	def Update(self):
		"""
		Updates the world's state every game tick.
		"""
		[e.Update() for e in self.entities]
		
		# check if our player has walked near the screen bounds and scroll over
		if self.player.scaledposition[0] > 0.2:
			self.position[0] += self.player.velocity
			self.player.Rescale()
		if self.player.scaledposition[0] < -0.2:
			self.position[0] -= self.player.velocity
			self.player.Rescale()
		
		if self.player.position[1] > self.position[1] + 0.2:
			self.position[1] += self.player.velocity
		if self.player.position[1] < self.position[1]:
			self.position[1] -= self.player.velocity

		# see if we need to create new items
		if self.position[0] > self.lastcreated[0] + 0.25:
			self.lastcreated[0] += 0.25
			for y in range(-8, 4):
				self.CreateStuff(self.lastcreated[0] + 1.0, self.lastcreated[1] + y/4.0)

		if self.position[0] < self.lastcreated[0] - 0.25:
			self.lastcreated[0] -= 0.25
			for y in range(-8, 4):
				self.CreateStuff(self.lastcreated[0] - 1.0, self.lastcreated[1] + y/4.0)

		if self.position[1] > self.lastcreated[1] + 0.25:
			self.lastcreated[1] += 0.25
			for x in range(-4, 4):
				self.CreateStuff(self.lastcreated[0] + x/4.0, self.lastcreated[1] + 1.0)

		if self.position[1] < self.lastcreated[1] - 0.25:
			self.lastcreated[1] -= 0.25
			for x in range(-8, 8):
				self.CreateStuff(self.lastcreated[0] + x/4.0, self.lastcreated[1] - 2.0)
		
		# check if any entities have falled by the wayside
		# delete them if they're too far away for us to see
		[self.RemoveEntity(e) for e in self.entities if self.Distance(self.player.position, e.position) > 2.5]
		
		self.DetectCollisions()

	
	def Draw(self):
		"""
		Draw everything in the world, every game tick.
		"""
		# sort by Y value (closeness to us)
		self.entities.sort(self.SortByY)
		# draw everything
		[e.Draw() for e in self.entities]
		# draw the hud
		self.hud.Draw()
	
	def DetectCollisions(self):
		"""
		Test for collisions between the player and all entities.
		"""
		# forget everything that was near us
		self.player.ForgetNearStuff()
		# find things (Scenery) we have run into
		[self.player.Collide(e) for e in self.entities if (isinstance(e, Scenery) and self.Distance(self.player.position, e.position) < e.width * e.scale / 2.0)]
		# find things (Items) on the ground that we are standing near
		[self.player.NearItem(e) for e in self.entities if (isinstance(e, Item) and self.Distance(self.player.position, e.position) < 0.1)]
		[(self.player.Fight(e) or e.Fight(self.player)) for e in self.entities if (isinstance(e, Character) and (e != self.player) and self.Distance(self.player.position, e.position) < 0.1)]
		[e.NearPlayer() for e in self.entities if (isinstance(e, Character) and (e != self.player) and self.Distance(self.player.position, e.position) < 1.0)]
	
	def PickUp(self, items):
		"""
		The player has picked up these items, so we need to remove them from the world.
		"""
		# make sure they don't get deleted and stay in the weakref.
		self.pickedup.extend(items)
		[self.entities.remove(i) for i in items]
	
	def Distance(self, a, b):
		"""
		Calculate the distance between two vectors.
		"""
		return sqrt(pow(a[0] - b[0], 2) + pow(a[1] - b[1], 2))
	
	def SortByY(self, a, b):
		return cmp(a.position[1], b.position[1])
	
	def Scale(self, x, y):
		"""
		Tell our children what scale they should be at, depending on where they are.
		Returns a scale and position depending upon the input scale and position.
		In this case as the player moves away into the distances we want to make them smaller.
		We want to move them slowly upward and then downward again as we move further away too.
		This is pseudo 3d with an animal-crossing style viewpoint.
		"""
		# make these guys relative to the world itself
		x = x - self.position[0]
		y = y - self.position[1]
		
		# y maps to a hill - as you walk away from the screen you go up and then down slightly
		#y = -(y - 1.0) * (y - 1.0) * (log(y) + e) / 2.0 + 1.0
		
		scale = (y + self.horizon) / self.horizon
		x = x * (y + self.horizon) / self.horizon
		y = sin(y / 1.5 - 0.75) * 0.9 + 0.5
		return scale, [x, y]
	
